/// HEADER
#include "background_subtraction.h"

/// COMPONENT
#include "adaptive_background_remover.h"
#include "opencv_background_remover.h"
#include "simple_background_remover.h"

/// PROJECT
#include <csapex/msg/io.h>
#include <utils_param/parameter_factory.h>
#include <csapex/model/node_modifier.h>
#include <csapex/utility/register_apex_plugin.h>

/// SYSTEM
#include <boost/assign/std.hpp>
#include <boost/lambda/lambda.hpp>

CSAPEX_REGISTER_CLASS(csapex::BackgroundSubtraction, csapex::Node)

using namespace csapex;
using namespace background_subtraction;
using namespace connection_types;

using namespace boost::lambda;

BackgroundSubtraction::BackgroundSubtraction()
{
}

void BackgroundSubtraction::setupParameters(Parameterizable& parameters)
{
    std::map<std::string, int> methods = boost::assign::map_list_of
            ("OpenCV", (int) OpenCV)
            ("Simple", (int) Simple)
            ("Adaptive", (int) Adaptive);

    std::function<void(param::Parameter*)> reset = std::bind(&BackgroundSubtraction::resetRemover, this);
    std::function<void(param::Parameter*)> update = std::bind(&BackgroundSubtraction::updateRemover, this);

    param::Parameter::Ptr method = param::ParameterFactory::declareParameterSet("method", methods, (int) OpenCV);
    parameters.addParameter(method, reset);

    parameters.addParameter(param::ParameterFactory::declareRange("threshold", 0, 100, 20, 1), update);
    parameters.addParameter(param::ParameterFactory::declareRange("open", 0, 20, 0, 1), update);
    parameters.addParameter(param::ParameterFactory::declareRange("close", 0, 20, 0, 1), update);

    /// OPENCV
    std::function<bool()> cond_opencv = [method]() { return method->as<int>() == OpenCV; };
    parameters.addConditionalParameter(param::ParameterFactory::declareRange("rate", 0.0, 0.1, 0.002, 0.001),
                            cond_opencv, update);
    parameters.addConditionalParameter(param::ParameterFactory::declareRange("history", 0, 256, 16, 1),
                            cond_opencv, update);

    /// ADAPTIVE
    std::function<bool()> cond_adaptive = [method]() { return method->as<int>() == Adaptive; };
    parameters.addConditionalParameter(param::ParameterFactory::declareRange("max. distance", 0.0, 50.0, 4.0, 0.1),
                            cond_adaptive, update);
    parameters.addConditionalParameter(param::ParameterFactory::declareRange("max. std. dev.", 0.0, 50.0, 4.0, 0.1),
                            cond_adaptive, update);
    parameters.addConditionalParameter(param::ParameterFactory::declareRange("decay", 0.0, 1.0, 0.7, 0.01),
                            cond_adaptive, update);


    parameters.addParameter(param::ParameterFactory::declareTrigger("Set Background"),
                 std::bind(&BackgroundSubtraction::setBackground, this));
}

void BackgroundSubtraction::setup(NodeModifier& node_modifier)
{
    input_ = node_modifier.addInput<CvMatMessage>("Image");

    out_masked_ = node_modifier.addOutput<CvMatMessage>("Masked Image");
    out_mask_ = node_modifier.addOutput<CvMatMessage>("Mask");
    out_bg_ = node_modifier.addOutput<CvMatMessage>("Background");
}

void BackgroundSubtraction::resetRemover()
{
    int type = readParameter<int>("method");

    {
        std::lock_guard<std::mutex> lock(remover_mutex);
        switch(type) {
        case OpenCV:
        {
            std::shared_ptr<OpenCvBackgroundRemover> ptr(new OpenCvBackgroundRemover);
            remover = ptr;
        }
            break;
        case Simple:
        {
            std::shared_ptr<SimpleBackgroundRemover> ptr(new SimpleBackgroundRemover);
            remover = ptr;
        }
            break;
        case Adaptive:
        {
            std::shared_ptr<AdaptiveBackgroundRemover> ptr(new AdaptiveBackgroundRemover);
            remover = ptr;
        }
            break;
        }
    }

    updateRemover();
    setBackground();
}

void BackgroundSubtraction::updateRemover()
{
    if(!remover) {
        return;
    }

    int type = readParameter<int>("method");

    {
        std::lock_guard<std::mutex> lock(remover_mutex);
        switch(type) {
        case OpenCV:
        {
            std::shared_ptr<OpenCvBackgroundRemover> ptr = std::dynamic_pointer_cast<OpenCvBackgroundRemover>(remover);
            ptr->setHistory(readParameter<int>("history"));
            ptr->setRate(readParameter<double>("rate"));
        }
            break;

        case Adaptive:
        {
            std::shared_ptr<AdaptiveBackgroundRemover> ptr = std::dynamic_pointer_cast<AdaptiveBackgroundRemover>(remover);
            ptr->setMaxDistance(readParameter<double>("max. distance"));
            ptr->setMaxStdDev(readParameter<double>("max. std. dev."));
            ptr->setDecay(readParameter<double>("decay"));
        }
            break;
        }

        remover->setOpen(readParameter<int>("open"));
        remover->setClose(readParameter<int>("close"));
        remover->setThreshold(readParameter<int>("threshold"));
    }
}

void BackgroundSubtraction::process()
{
    if(!remover) {
        resetRemover();
    }

    connection_types::CvMatMessage::ConstPtr frame = msg::getMessage<CvMatMessage>(input_);
    assert(frame);

    connection_types::CvMatMessage::Ptr filtered(new connection_types::CvMatMessage(frame->getEncoding(), frame->stamp_micro_seconds));
    connection_types::CvMatMessage::Ptr mask(new connection_types::CvMatMessage(enc::mono, frame->stamp_micro_seconds));
    {
        std::lock_guard<std::mutex> lock(remover_mutex);
        if(!bg_frame) {
            bg_frame.reset(new connection_types::CvMatMessage(enc::mono, frame->stamp_micro_seconds));
            frame->value.copyTo(bg_frame->value);

            remover->setBackground(bg_frame->value);
        }

        remover->filter(frame->value, filtered->value, mask->value);

        if(msg::isConnected(out_bg_)) {
            connection_types::CvMatMessage::Ptr bg(new connection_types::CvMatMessage(frame->getEncoding(), frame->stamp_micro_seconds));
            remover->getBackground().copyTo(bg->value);
            msg::publish(out_bg_, bg);
        }
    }

    msg::publish(out_masked_, filtered);
    msg::publish(out_mask_, mask);
}

void BackgroundSubtraction::setBackground()
{
    if(bg_frame) {
        bg_frame.reset((connection_types::CvMatMessage*) nullptr);
    }
}
