#include "imagereprojector.h"
#include <algorithm>
#include <Eigen/Core>
#include <utils_general/MathHelper.h>

using namespace std;
using namespace Eigen;


void ImageReprojector::computePose(double px, double py, const lib_laser_processing::Scan &scan, const tf::Transform &scan_to_camera,
                                   tf::Vector3 *pos, tf::Quaternion *rot, int num_steps)
{
    double x_angle = std::atan2(px, 1);

    double z = getDistance(x_angle, scan, scan_to_camera);
    double x = px * z;
    double y = py * z;

    *pos = tf::Vector3(x,y,z);

    if (rot != NULL) {
        *rot = computeRotation(x_angle, scan, num_steps);
    }
}

tf::Quaternion ImageReprojector::computeRotation(double angle, const lib_laser_processing::Scan &scan, int num_steps)
{
    // sample some scan points around the target
    size_t ray_index = (angle - scan.angle_min) / scan.angle_increment;
    vector<Vector2d> points;
    for (size_t i = ray_index - num_steps; i <= ray_index + num_steps; ++i) {
        if(std::abs(scan.rays[i].range - scan.rays[ray_index].range) < 1.0) {
            double angle = scan.angle_min + i * scan.angle_increment;
            Vector2d p = scan.rays[i].range * Vector2d(std::cos(angle), std::sin(angle));
            points.push_back(p);
        }
    }

    MathHelper::Line line = MathHelper::FitLinear(points);
    //line has to go left to right (= positive on y-axis)
    if (line.direction[1] < 0) {
        line.direction *= -1;
    }
    tf::Vector3 normal(line.direction[1], line.direction[0], 0);
    double target_yaw = std::atan2(normal.y(), normal.x());

    return tf::createQuaternionFromYaw(target_yaw);
}

float ImageReprojector::getDistance(float scan_angle, const lib_laser_processing::Scan &scan,
                                    const tf::Transform &scan_to_camera)
{
    int index = (scan_angle - scan.angle_min) / scan.angle_increment;

    //TODO: out of bounds check

    int steps = 3;//TODO: is 3 a good number?

    // compute median
    vector<float> ranges;
    ranges.reserve(2*steps+1);
    for (int i = index-steps; i <= index+steps+1; ++i) {
        //ranges.push_back( scan.rays[i].range );

        tf::Point point_laser;
        point_laser.setX(scan.rays[i].range * std::cos(scan_angle));
        point_laser.setY(scan.rays[i].range * std::sin(scan_angle));
        tf::Point point_base = scan_to_camera * point_laser;

        //ranges.push_back( point_base.length() );
        ranges.push_back( point_base.getZ());
    }
    sort(ranges.begin(), ranges.end());
    float median = ranges[ranges.size()/2];

    return median;
}
