#ifndef IMAGEREPROJECTOR_H
#define IMAGEREPROJECTOR_H

#include <tf/tf.h>
#include <utils_laser_processing/data/scan.h>

/**
 * @brief Reproject points from image plane to camera frame poses.
 *
 * Used for signs and targets. This class assumes, that the reprojected point is on a flat wall, as laser scans are used
 * to estimate distance and orientation.
 * Orientation is computed by fitting a line to the laser scan points around the transformed point (-> flat wall needed!).
 *
 * @todo: move undistortion of points here to get full pixel to pose transformation?
 */
class ImageReprojector
{
public:
    /**
     * @brief Reproject a point from the image to the camera frame. Laser scan is used for distance and optionally
     *        to estimate orientation.
     * @param px        x-ccordinate of the point on the image plane.
     * @param py        y-coordinate of the point on the image plane.
     * @param scan      Laser scan.
     * @param pos       [out] Position in camera frame.
     * @param rot       [out] Orientation (only comupted if != NULL)
     * @param num_steps Number of used scan points left and right of the object (total used number is 2*num_steps+1)
     * @see computeRotation
     */
    static void computePose(double px, double py, const lib_laser_processing::Scan &scan, const tf::Transform &scan_to_camera,
                            tf::Vector3 *pos, tf::Quaternion *rot=NULL, int num_steps=15);

    /**
     * @brief Compute orientation of a sign/target, assuming that it is fixed on a flat wall.
     * @param angle     Scan angle to the object.
     * @param scan      Laser scan
     * @param num_steps Number of used scan points left and right of the object (total used number is 2*num_steps+1)
     * @return The computed orientation as a quaternion in the frame of the laser scanner.
     */
    static tf::Quaternion computeRotation(double angle, const lib_laser_processing::Scan &scan, int num_steps=15);

private:
    /**
     * @brief Get distance from laser scan for the given angle.
     * @return Distance
     */
    static float getDistance(float scan_angle, const lib_laser_processing::Scan &scan,
                             const tf::Transform &scan_to_camer);
};

#endif // IMAGEREPROJECTOR_H
