#include "mask_image_from_scan_node.h"

/// CSAPEX INTERFACES
#include <csapex/msg/io.h>
#include <utils_param/parameter_factory.h>
#include <csapex/model/node_modifier.h>

/// MESSAGES
#include <csapex_scan_2d/scan_message.h>
#include <csapex_vision/cv_mat_message.h>

/// SYSTEM
#include <csapex/utility/register_apex_plugin.h>
#include <Eigen/Core>

CSAPEX_REGISTER_CLASS(csapex::MaskImageFromScanNode, csapex::Node)

using namespace csapex;
using namespace csapex::connection_types;
using namespace lib_laser_processing;


void MaskImageFromScanNode::loadCameraParameterFromFile()
{
    /// Setting default values
    camera_matrix_.at<double>(0,0) = 7.5635817817850432e+02;
    camera_matrix_.at<double>(0,1) = 0.0;
    camera_matrix_.at<double>(0,2) = 2.3995466710184999e+02;

    camera_matrix_.at<double>(1,0) = 0.0;
    camera_matrix_.at<double>(1,1) = 7.4430004898568041e+02;
    camera_matrix_.at<double>(1,2) = 2.9763314181994497e+02;

    camera_matrix_.at<double>(2,0) = 0.0;
    camera_matrix_.at<double>(2,1) = 0.0;
    camera_matrix_.at<double>(2,2) = 1.0;

    // Create zero distortion
    distortion_coeffs_.at<double>(0) = 0;
    distortion_coeffs_.at<double>(1) = 0;
    distortion_coeffs_.at<double>(2) = 0;
    distortion_coeffs_.at<double>(3) = 0;


    std::string path = readParameter<std::string>("camera parameter");
    if(path == "") {
        aerr << "Could not load camera parameter. Using default...." << std::endl;
    }

    try {
        cv::FileStorage fs(path, cv::FileStorage::READ);
        if(!fs.isOpened()) {
            aerr << "cannot open config at " << path << std::endl;
            return;
        }
        fs["intrinsics"] >> camera_matrix_;
        fs["distortion"] >> distortion_coeffs_;
        fs.release();
        ainfo << "Loaded camera parameter\ncamera_matrix:\n" << camera_matrix_
              << "\ndistoriton_coeffs_:\n" <<distortion_coeffs_ << std::endl;
    } catch(const cv::Exception& e) {
        aerr << "Could not open file. Using default parameter ..." << std::endl;
    }
}



void MaskImageFromScanNode::createMaskFromLaser(const lib_laser_processing::Scan& scan,
                         const tf::Transform& tf,
                         const cv::Mat& img_in,
                         cv::Mat &img_debug,
                         cv::Mat &mask,
                         double par_wall_height,
                         double par_laser_height,
                         double par_laser_opening_angle)
{
    mask = cv::Mat (img_in.rows, img_in.cols, CV_8UC1, cv::Scalar::all(0));

   // Example found here: https://github.com/daviddoria/Examples/blob/master/c%2B%2B/OpenCV/ProjectPoints/ProjectPoints.cxx

   /// --------- Generate 3D Points -------
   /// ------------------------------------

   // Convert 2d Laser Points to 3D and filter by range and angle
   std::vector<cv::Point3d> laser_points_3d;
   for (unsigned int i = 0, total = scan.rays.size(); i < total; i++) {
       const LaserBeam& beam = scan.rays.at(i);
       if (beam.range < scan.range_max && beam.range > scan.range_min && beam.yaw < par_laser_opening_angle/2 && beam.yaw > -par_laser_opening_angle/2) {

           // convert point to x,y,z
           cv::Point3d cv_point_laser;
           cv_point_laser.x =  beam.pos(0);
           cv_point_laser.y =  beam.pos(1);
           cv_point_laser.z =  0;
           laser_points_3d.push_back(cv_point_laser);
       }
   }


   // Generate 3D mask points from laser points
   std::vector<cv::Point3d> mask_points_3d;
   mask_points_3d = laser_points_3d;

   // Add bottom of mask (wall bottom edege)
   for (unsigned int i = 0, total = laser_points_3d.size(); i < total; i++) {
       mask_points_3d.at(i).z = par_laser_height;
   }
   // Add top of mask (wall top edge)
   for (int i = laser_points_3d.size()-1; i >=0 ; i--) {
       cv::Point3d cv_point_laser;
       cv_point_laser.x =  laser_points_3d.at(i).x;
       cv_point_laser.y =  laser_points_3d.at(i).y;
       cv_point_laser.z =  par_laser_height - par_wall_height;
       mask_points_3d.push_back(cv_point_laser);
   }


   /// --------- Setup projection prameters -------
   /// ------------------------------------------

   // see: CV rotation matrix: http://docs.opencv.org/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html#void%20Rodrigues%28InputArray%20src,%20OutputArray%20dst,%20OutputArray%20jacobian%29
   // see: TF rotation matrix chapter 3.: http://wiki.ros.org/tf/Overview/Transformations
  cv::Mat rvec(3,3,cv::DataType<double>::type); // rotation matrix
  tf::Matrix3x3 tf_rotation =  tf.getBasis();
  rvec.at<double>(0,0) = tf_rotation[0][0];
  rvec.at<double>(0,1) = tf_rotation[0][1];
  rvec.at<double>(0,2) = tf_rotation[0][2];

  rvec.at<double>(1,0) = tf_rotation[1][0];
  rvec.at<double>(1,1) = tf_rotation[1][1];
  rvec.at<double>(1,2) = tf_rotation[1][2];

  rvec.at<double>(2,0) = tf_rotation[2][0];
  rvec.at<double>(2,1) = tf_rotation[2][1];
  rvec.at<double>(2,2) = tf_rotation[2][2];

  cv::Mat T(3,1,cv::DataType<double>::type); // translation vector
  tf::Vector3 tf_translation = tf.getOrigin();
  T.at<double>(0,0) = tf_translation[0];
  T.at<double>(1,0) = tf_translation[1];
  T.at<double>(2,0) = tf_translation[2];

  cv::Mat rvecR(3,1,cv::DataType<double>::type); //rodrigues rotation matrix
  cv::Rodrigues(rvec,rvecR);


  /// --------- Project Points to image  -------
  /// ------------------------------------------

  std::vector<cv::Point2d> laser_points_projected;
  if (laser_points_3d.size() > 0) {
    cv::projectPoints(laser_points_3d, rvecR, T, camera_matrix_, distortion_coeffs_, laser_points_projected);
   } else {
      aerr << "No Valid Laser points to Project (received " << scan.rays.size() << " points)" << std::endl;
  }


  std::vector<cv::Point2d> mask_points_projected;
  if (mask_points_3d.size() > 0) {
    cv::projectPoints(mask_points_3d, rvecR, T, camera_matrix_, distortion_coeffs_, mask_points_projected);
   } else {
      aerr << "No valid mask points to project (received " << scan.rays.size() << " points)" << std::endl;
  }


  /// --------- Draw projected points to images  -------
  /// ------------------------------------------

   img_debug.create(img_in.rows, img_in.cols, CV_8UC3);
//   img_in.convertTo(img_debug, CV_8UC3);
   assert(img_debug.type() == CV_8UC3);
   // Draw laser_points_projected to the img_debug
   int number_of_laser_points_outside_image = 0;
   for (unsigned int i = 0, total = laser_points_projected.size(); i < total; i++) {
       const cv::Point2d& point = laser_points_projected.at(i);
       int y = (int)round(point.y);
       int x = (int)round(point.x);
       if ( x >= 0 && x < img_debug.cols && y >= 0 && y < img_debug.rows) {
           cv::Vec3b& pixel = img_debug.at<cv::Vec3b>(y, x);
           pixel = cv::Vec3b(0,i%255,255);
       } else {
           number_of_laser_points_outside_image++;
       }
   }
   // awarn << "Lost " << number_of_laser_points_outside_image << " of " << laser_points_projected.size() << " laser points that do not fit on the image " << std::endl;


   /// Generate the points for the mask and check how many are out of the image
   unsigned number_of_mask_points_outside_image = 0;
   std::vector<cv::Point> image_points;
   for (std::vector<cv::Point2d>::iterator it = mask_points_projected.begin(); it != mask_points_projected.end();it++) {
       const cv::Point2d& point = *it;
       int y = (int) std::floor(point.y + 0.5);
       int x = (int) std::floor(point.x + 0.5);

       // Check how many mask points are outside the image
       if ( x >= 0 && x < img_debug.cols &&
            y >= 0 && y < img_debug.rows) {
           // Draw mask points to debug image
           cv::Vec3b& pixel = img_debug.at<cv::Vec3b>(y, x);
           pixel = cv::Vec3b(0,255,255);
       } else {
           ++number_of_mask_points_outside_image;
       }

       // Fit mask points to mask image
       x = std::min(x, img_debug.cols-1); // limit to coll size
       x = std::max(0, x);
       y = std::min(y, img_debug.rows-1);
       y = std::max(0, y);
       image_points.push_back(cv::Point(x,y));
   }
   // awarn << "Lost " << number_of_mask_points_outside_image << " of " << mask_points_projected.size()  << " mask points that do not fit on the image " << std::endl;

   // If all projected points are outside the image (for verry close objects)
   if (image_points.size() < 2) {
       // set the whole mask image to black
       mask.setTo(cv::Scalar::all(0));
   } else if (number_of_mask_points_outside_image == mask_points_projected.size()) {
       // set the whole mask image to white
       mask.setTo(cv::Scalar::all(255));
   } else {
       // draw polygon to mask image

       std::vector< std::vector<cv::Point> > image_points_vector;
       image_points_vector.push_back(image_points);
       cv::fillPoly(mask,image_points_vector,cv::Scalar::all(255));
   }
}

MaskImageFromScanNode::MaskImageFromScanNode()
{
    camera_matrix_ = cv::Mat(3,3,cv::DataType<double>::type);
    distortion_coeffs_ = cv::Mat(4,1,cv::DataType<double>::type);
}


void MaskImageFromScanNode::process()
{
    /// Get the Input Data
    ScanMessage::ConstPtr scan_msg = msg::getMessage<ScanMessage>(input_scan_);
    CvMatMessage::ConstPtr image_msg = msg::getMessage<CvMatMessage>(input_image_);
    TransformMessage::ConstPtr trafo_msg = msg::getMessage<TransformMessage>(input_transform_);

    /// Outputs
    CvMatMessage::Ptr debug(new CvMatMessage(enc::bgr, image_msg->stamp_micro_seconds));
    CvMatMessage::Ptr mask_msg(new CvMatMessage(enc::mono, image_msg->stamp_micro_seconds));

    /// Do some work here
    createMaskFromLaser(scan_msg->value, trafo_msg->value,
                        image_msg->value, debug->value, mask_msg->value,
                readParameter<double>("wall height"),
                readParameter<double>("laser height"),
                readParameter<double>("laser opening angle"));


    /// Publish output Data:
    msg::publish(output_debug_, debug);
    msg::publish(output_mask_, mask_msg);
}

void MaskImageFromScanNode::setup(csapex::NodeModifier& node_modifier)
{
    input_scan_      = modifier_->addInput<ScanMessage>("Scan");
    input_image_     = modifier_->addInput<CvMatMessage>("Image");
    input_transform_ =  modifier_->addInput<connection_types::TransformMessage>("T");

    output_mask_     = modifier_->addOutput<CvMatMessage>("Mask");
    output_debug_    = modifier_->addOutput<CvMatMessage>("Debug");
}

void MaskImageFromScanNode::setupParameters(Parameterizable& parameters)
{
    addParameter(param::ParameterFactory::declareRange<double>("wall height", 0, 2, 0.5, 0.01));
    addParameter(param::ParameterFactory::declareRange<double>("laser height", -2, 2, 0.13, 0.01));
    addParameter(param::ParameterFactory::declareRange<double>("laser opening angle", 0, 6.3, 0.7, 0.01));
    addParameter(param::ParameterFactory::declareFileInputPath("camera parameter", ""), boost::bind(&MaskImageFromScanNode::loadCameraParameterFromFile, this));
}


