#include "cross_detection.h"

/// PROJECT
#include <csapex/msg/io.h>
#include <csapex/model/node_modifier.h>
#include <utils_param/parameter_factory.h>
#include <sick14/cross_message.h>

/// SYSTEM
#include <csapex/utility/register_apex_plugin.h>
#include <boost/assign/std.hpp>

CSAPEX_REGISTER_CLASS(csapex::CrossDetection, csapex::Node)

using namespace csapex;
using namespace connection_types;


CrossDetection::CrossDetection()
{
    addParameter(param::ParameterFactory::declareRange("rho", 1.0, 100.0, 1.0, 1.0),
                 boost::bind(&CrossDetection::update, this));
    addParameter(param::ParameterFactory::declareRange("theta", 0.1, 2 * CV_PI, CV_PI, 0.1),
                 boost::bind(&CrossDetection::update, this));
    addParameter(param::ParameterFactory::declareRange("threshold", 1, 200, 100, 1),
                 boost::bind(&CrossDetection::update, this));
    addParameter(param::ParameterFactory::declareRange("min length ratio", 0.0, 1.0, 0.25, 0.05),
                 boost::bind(&CrossDetection::update, this));
    addParameter(param::ParameterFactory::declareRange("max gap ratio", 0.0, 1.1, 0.3, 0.05),
                 boost::bind(&CrossDetection::update, this));
    addParameter(param::ParameterFactory::declareRange("angle tolerance", 1, 50, 5, 1),
                 boost::bind(&CrossDetection::update, this));
}


void CrossDetection::setup(csapex::NodeModifier& node_modifier)
{
    in_mat_ = modifier_->addInput<CvMatMessage>("Original Image");
    in_rois_ = modifier_->addInput<VectorMessage, RoiMessage>("ROIs");
    in_edges_ = modifier_->addInput<CvMatMessage>("Edges");


    out_debug_ = modifier_->addOutput<CvMatMessage>("OutputImage");
    out_cross_ = modifier_->addOutput<CrossMessage>("Target Cross");
    update();

}

void CrossDetection::update()
{
    // Hough lines parameter
    rho_ = readParameter<double>("rho");
    theta_ = readParameter<double>("theta");
    threshold_ = readParameter<int>("threshold");
    line_ratio_ = readParameter<double>("min length ratio");
    gap_ratio_ = readParameter<double>("max gap ratio");

    tol_ = readParameter<int>("angle tolerance");
}

void CrossDetection::detectTarget(const cv::Mat &img, u_int64_t stamp, const cv::Mat &edges,
                                  std::vector<Result> &results)
{
    CvMatMessage::Ptr out(new connection_types::CvMatMessage(enc::bgr, stamp));
    cv::cvtColor(img, out->value, CV_GRAY2BGR);

    // ROI-Handling
    VectorMessage::ConstPtr rois = msg::getMessage<VectorMessage>(in_rois_);

    for(std::vector<ConnectionType::ConstPtr>::const_iterator it = rois->value.begin(); it != rois->value.end(); ++it)
    {
        RoiMessage::ConstPtr roi = std::dynamic_pointer_cast<RoiMessage const>(*it);

        cv::Mat curr_roi;
        cv::Mat(img, roi->value.rect()).copyTo(curr_roi);

        cv::Mat edges_roi;
        cv::Mat(edges, roi->value.rect()).copyTo(edges_roi);

        // scale the minimal length  and max gap with the rois size
        int min_line_length = line_ratio_ * std::min(roi->value.h(), roi->value.w());
        int max_line_gap = gap_ratio_ * std::min(roi->value.h(), roi->value.w());

        std::vector<cv::Vec4i> lines;
        cv::HoughLinesP(edges_roi, lines, rho_, theta_/180, threshold_, min_line_length, max_line_gap);


        filterLines(out->value, lines, roi->value.x(), roi->value.y(), roi->value.w(), roi->value.h(), out->value.cols, out->value.rows);


        Result res = detectCross(out->value, lines, roi->value);

        if(!res.found)
        {
            detectLine(out->value, lines, roi->value, res);
        }
        results.push_back(res);

    }
    msg::publish(out_debug_, out);

}

CrossDetection::Result CrossDetection::detectCross(cv::Mat &out,const std::vector<cv::Vec4i> &lines, const Roi &roi)
{
    std::vector<cv::Point2i> target_centers;
    for( size_t i = 0; i < lines.size(); i++ ) {
        cv::Point2i o_start = cv::Point(lines[i][0]+  roi.x(), lines[i][1] +  roi.y());
        cv::Point2i o_end = cv::Point(lines[i][2] +  roi.x(), lines[i][3]+  roi.y());

        for (size_t j = i +1; j < lines.size(); j++ )
        {
            cv::Point2i p_start = cv::Point(lines[j][0] +  roi.x(), lines[j][1]+  roi.y());
            cv::Point2i p_end = cv::Point(lines[j][2] +  roi.x(), lines[j][3]+  roi.y());

            // skip pairs of lines with too different lengths
            double ratio = cv::norm(p_start-p_end)/cv::norm(o_start-o_end);
            if (ratio > 4./3. || ratio < 0.75)
            {
                //                    std::cout << "skipped lines with length-ratio " << ratio << std::endl;
                continue;
            }

            // check for intersections
            double dyo = o_end.y - o_start.y;
            double dxo = o_start.x-o_end.x;
            double co = dyo*o_start.x + dxo *o_start.y;
            double dyp = p_end.y - p_start.y;
            double dxp = p_start.x-p_end.x;
            double cp = dyp*p_start.x + dxp *p_start.y;

            float cross = dyo*dxp - dyp*dxo;
            if (abs(cross) < 1e-5)
                continue;

            // check angle between lines, should be 90°
            double phi = std::acos((dxo*dxp+dyo*dyp)/(sqrt(dxo*dxo+dyo*dyo)*sqrt(dxp*dxp+dyp*dyp)));
            if (phi*180/CV_PI < 80){
                continue;
            }


            double x = (dxp*co - dxo*cp)/cross;
            double y = (dyo*cp - dyp*co)/cross;
            cv::Point2i r = cv::Point2d(round(x),round(y));

            // check if the intersection-point lies inside the roi and far enough from the rois edges
            // if the roi is properly contained in the image
            /// TODO: adjust if region finder is not reliable
            int w_margin = roi.w() /3;
            int h_margin = roi.h() /3;

            if ((r.x >= roi.x() + w_margin || (roi.x() <= 5 && r.x >= 10))
                    && (r.x <= roi.x() +roi.w() - w_margin
                        || (roi.x() + roi.w() >= out.cols - 5 && r.x < out.cols - 11))
                    && (r.y >= roi.y() + h_margin || (roi.y() <= 5 && r.y >= 10))
                    && (r.y <= roi.y() +roi.h() - h_margin
                        || (roi.y() + roi.h() >= out.rows - 5 && r.y < out.rows - 11)))
            {
                target_centers.push_back(r);
                cv::circle(out, r, 50, cv::Scalar(255,0,255) );
            }
        }
    }

    Result res;
    res.roi_size = roi.h()*roi.w();
    res.center = cv::Point2i(roi.x()+ roi.w()/2, roi.y() + roi.h()/2);
    res.roi_intersects = roi.x() < 5 || roi.x() + roi.w() > out.cols - 6 ||
            roi.y() < 5 || roi.y() + roi.h() > out.rows - 6;
    res.found = false;
    res.cross = false;

    // check possible target centers
    if (target_centers.size() > 0)
    {
        cv::Point2i avg;
        for (size_t  i = 0; i < target_centers.size(); i++)
        {
            avg += target_centers[i];
        }
        avg.x = avg.x / target_centers.size();
        avg.y = avg.y / target_centers.size();

        int count = 0;
        int avg_x = 0;
        int avg_y = 0;
        for (size_t i = 0; i < target_centers.size(); i++)
        {
            cv::Point2i t = target_centers[i];
            /// TODO threshold
            if (cv::norm(t-avg) < roi.h()/10)
            {
                count++;
                avg_x += t.x;
                avg_y += t.y;
            }
        }

        /// TODO threshold (kann man evtl höher setzten? kommt auf entfernung an
        if (count >= std::max(1,(int) target_centers.size()/5))
        {
            res.point = cv::Point2i(avg_x/count, avg_y/count);
            res.found = true;
            res.cross = true;
            cv::circle(out, cv::Point2i(avg_x/count, avg_y/count), 5, cv::Scalar(255,0,0), 10 );
        }
    }
    return res;
}

void CrossDetection::detectLine(cv::Mat &out, const std::vector<cv::Vec4i> &lines, const Roi &roi, CrossDetection::Result &res)
{
    std::vector<cv::Point2i> points;

    for( size_t i = 0; i < lines.size(); i++ ) {
        // calculate start and end point in image-coordinates
        cv::Point2i start = cv::Point(lines[i][0]+  roi.x(), lines[i][1] +  roi.y());
        cv::Point2i end = cv::Point(lines[i][2] +  roi.x(), lines[i][3]+  roi.y());


        // reject lines that aren't vertical enough
        if (abs(start.x - end.x) > tol_ )
        {
            continue;
        }

        points.push_back(start);
    }


    // check the lines
    if (points.size() == 1)
    {
        // not much evidence here, let's hope its ok
        res.found = true;
        res.point = points.front();

        cv::circle(out, points.front(), 5, cv::Scalar(255,0,0), 10 );
    } else {
        // search for pairs of (parallel) lines that are close enough
        /// TODO Wert
        int max_dist = 50;
        int min_dist = 3;
        std::vector<int> avg_x;
        std::vector<int> y;
        std::vector<int> dists;

        for (std::vector<cv::Point2i>::iterator it = points.begin(); it != points.end(); ++it)
        {
            int x_1 = it->x;
            int y_1 = it->y;
            for (std::vector<cv::Point2i>::iterator it2 = it+1; it2 != points.end(); ++it2)
            {
                int x_2 = it2->x;

                if (abs(x_1-x_2) < max_dist && abs(x_1-x_2) > min_dist)
                {
                    dists.push_back(abs(x_1-x_2));
                    avg_x.push_back((x_1 + x_2)/2);
                    y.push_back(y_1);
                }
            }
        }

        // if we found some nice parallel lines
        if (avg_x.size() == 1)
        {
            res.point = (cv::Point2i(avg_x.front(), y.front()));
            res.found = true;

            cv::circle(out, cv::Point2i(avg_x.front(),y.front()), 5, cv::Scalar(255,0,0), 10 );
        }
        else if (avg_x.size() > 0)// ... panic?
        {
            int sum_x = 0;
            for (size_t i = 0; i < avg_x.size(); i++)
            {
                sum_x += avg_x[i];
            }

            int avg = sum_x/avg_x.size();

            int count = 0;
            int res_x = 0;
            int res_y = 0;
            for (size_t i = 0; i < avg_x.size(); i++)
            {
                if (abs(avg_x[i]-avg) < roi.w()/10)
                {
                    count++;
                    res_x += avg_x[i];
                    res_y = y[i];
                }
            }

            if (count > 2)
            {
                res.point = cv::Point2i(res_x/count, res_y);
                res.found = true;

                cv::circle(out, cv::Point2i(res_x/count, res_y), 5, cv::Scalar(255,0,0), 10 );
            }
        } else {
            // no line found
        }
    }
}

void CrossDetection::filterLines(cv::Mat &out,std::vector<cv::Vec4i> &lines, int r_x, int r_y, int r_w, int r_h, int w, int h)
{
    // check if roi's left or right border is the image's border
    bool left = r_x <= 3;
    bool top = r_y <= 3;
    bool right = r_x + r_w >= w -6;
    bool bottom = r_y + r_h >= h -6;

    for (std::vector<cv::Vec4i>::iterator it = lines.begin(); it != lines.end(); )
    {
        cv::Point2i p_start = cv::Point((*it)[0] +  r_x, (*it)[1]+  r_y);
        cv::Point2i p_end = cv::Point((*it)[2] +  r_x, (*it)[3]+  r_y);


        if (abs(p_start.x - p_end.x) > tol_ && abs(p_start.y - p_end.y) > tol_*2) // delete lines that aren't hoirzontal or vertical
        {
            // delete lines that are too close to the rois border
            it = lines.erase(it);

            cv::line(out, p_start,
                     p_end, cv::Scalar(255,0,0), 3, 8 );
        } else {
            bool horizontal = abs(p_start.y - p_end.y) < tol_*2; // if not, the line is vertical
            if (horizontal &&
                    ((!top && abs(p_start.y - r_y) < r_h/3) ||
                     (top && abs(p_start.y - r_y) < 10) ||
                     (!bottom && abs(p_start.y - (r_y + r_h)) < r_h/3) ||
                     (bottom && abs(p_start.y - (r_y + r_h)) < 10)))
            {
                // delete horizontal lines that are too close to the rois top or bottom
                it = lines.erase(it);

                cv::line(out, p_start,
                         p_end, cv::Scalar(255,0,0), 3, 8 );
            } else if (!horizontal &&
                       ((!left && abs(p_start.x - r_x) < r_w/3) ||
                        (left && abs(p_start.x - r_x) < 10) ||
                        (!right && abs(p_start.x - (r_x + r_w)) < r_h/3) ||
                        (right && abs(p_start.x - (r_x + r_w)) < 10)))
            {
                // delete horizontal lines that are too close to the rois top or bottom
                it = lines.erase(it);

                cv::line(out, p_start,
                         p_end, cv::Scalar(255,0,0), 3, 8 );
            } else {
                // keep line
                ++it;
                if (horizontal){
                    cv::line(out, p_start,
                             p_end, cv::Scalar(0,0,255), 3, 8 );
                } else {
                    cv::line(out, p_start,
                             p_end, cv::Scalar(0,255,0), 3, 8 );
                }
            }
        }
    }
}


void CrossDetection::process()
{
    CvMatMessage::ConstPtr msg = msg::getMessage<CvMatMessage>(in_mat_);
    CvMatMessage::ConstPtr edges = msg::getMessage<CvMatMessage>(in_edges_);

    if(!msg->hasChannels(1,CV_8U))  {
        throw std::runtime_error("image must be grayscale.");
    }

    std::vector<Result> results;

    detectTarget(msg->value, msg->stamp_micro_seconds, edges->value, results);


    int max_size_line = 0;
    int max_size_cross = 0;
    bool cross_found = false;
    bool line_found = false;
    CrossMessage::Ptr point(new CrossMessage(msg->frame_id, msg->stamp_micro_seconds));

    for (std::vector<Result>::iterator it = results.begin(); it != results.end(); ++it)
    {
        Result r = *it;
        if (r.found) //found something in this roi
        {
            if (r.cross && r.roi_size > max_size_cross)  //found a cross and the corresponding roi is the biggest so far
            {
                point->position.x = r.point.x;
                point->position.y = r.point.y;
                max_size_cross = r.roi_size;
                cross_found = true;

            } else if (! cross_found && r.roi_size > max_size_line) // found a line, haven't found a cross yet and the roi is the biggest so far
            {
                point->position.x = r.point.x;
                point->position.y = r.point.y;
                max_size_line = r.roi_size;
                line_found = true;
            }
        }
    }

    if (cross_found)
    {
        point->type = CrossMessage::INTERSECTION;
        //ainfo << "publish cross intersection" << std::endl;
        msg::publish(out_cross_, point);
    } else if (line_found)
    {
        point->type = CrossMessage::LINE;
        //ainfo << "do not publish cross line" << std::endl;
        //out_cross_->publish(point);
    }


}




