/// HEADER
#include "histogram_normalization.h"

/// PROJECT
#include <csapex/msg/io.h>
#include <csapex/model/node_modifier.h>
#include <utils_param/parameter_factory.h>

/// SYSTEM
#include <csapex/utility/register_apex_plugin.h>


CSAPEX_REGISTER_CLASS(csapex::HistogramNormalization, csapex::Node)

using namespace csapex;
using namespace connection_types;


HistogramNormalization::HistogramNormalization()
{
    addParameter(param::ParameterFactory::declareRange("threshold", 0, 400, 80, 1));
    addParameter(param::ParameterFactory::declareRange("number of bins", 10, 255, 50, 1));
}


void HistogramNormalization::setup(csapex::NodeModifier& node_modifier)
{
    in_mat_ = modifier_->addInput<CvMatMessage>("Original Image");
    in_mask_ = modifier_->addInput<CvMatMessage>("Mask");

    out_image_ = modifier_->addOutput<CvMatMessage>("Normalized");
    out_hist_ = modifier_->addOutput<CvMatMessage>("Original Histogram");

}

void HistogramNormalization::normalize(const cv::Mat &img, const cv::Mat &mask, u_int64_t stamp)
{
    int threshold = readParameter<int>("threshold");
    int histSize = readParameter<int>("number of bins");

    // Establish the size of the bins;
    double bin_size = 255./histSize;

    // Set the range
    float range[] = { 0, 255 } ;
    const float* histRange = { range };

    bool uniform = true;
    bool accumulate = false;

    cv::Mat hist;

    // Compute the histogram
    cv::calcHist( &img, 1, 0, mask, hist, 1, &histSize, &histRange, uniform, accumulate );

    int hist_w = 512; int hist_h = 400;
    int bin_w = cvRound( (double) hist_w/histSize );
    cv::Mat histImage( hist_h, hist_w, CV_8UC3, cv::Scalar( 0,0,0) );

    // Normalize the result to [ 0, histImage.rows ]
    cv::normalize(hist, hist, 0, histImage.rows, cv::NORM_MINMAX, -1);

    // Draw
    for( int i = 1; i < histSize; i++ )
    {
        cv::line( histImage, cv::Point( bin_w*(i-1), hist_h - cvRound(hist.at<float>(i-1)) ) ,
              cv::Point( bin_w*(i), hist_h - cvRound(hist.at<float>(i)) ),
              cv::Scalar( 255, 0, 0), 2, 8, 0  );
    }

    CvMatMessage::Ptr hist_out(new connection_types::CvMatMessage(enc::bgr, stamp));
    hist_out->value = histImage;
    msg::publish(out_hist_, hist_out);

    // find all bins above the threshold and the maximum of their corresponding peaks
    std::vector<int> val;
    std::vector<int> peak_width;

    for(int i = 0; i < histSize; ++i)
    {
        int bin  = cvRound(hist.at<float>(i));
        if (bin > threshold)
        {
            int max = bin;
            int max_ind = i;
            int width = 1;
            int next_bin = cvRound(hist.at<float>(++i));
            while (next_bin > threshold)
            {
                if (next_bin > max)
                {
                    max_ind = i;
                    max = next_bin;
                }
                next_bin = cvRound(hist.at<float>(++i));
                width++;
            }
            val.push_back(bin_size * max_ind + bin_size/2);
            peak_width.push_back(width*bin_size);
        }
    }


    // normalize the image
    double scale = 1;
    double shift = 0;
    double smin = 0;
    double min = val.front();
    double max = val.back();
    double smax = 0;

    cv::minMaxLoc(img, &smin, &smax, 0, 0, mask );


    std::cout << "max img: " << smax  << " max peak: " << max
              << " number of peaks: " << val.size() << std::endl;

   if (val.size() > 2 || (val.size() == 2 && min <= 50 )) {
       if (max - peak_width.back()/2 > max/2)
       {
           max = max - peak_width.back()/2;
       }
       std::cout << "used max-value: " << max << std::endl;
        scale = 255*((max - smin) > 0.001 ? 1./(max - smin) : 0);
        shift = - smin *scale;
    } else {
       scale = 255*((std::min(smax,190.) - smin) > 0.001 ? 1./(std::min(smax,190.) - smin) : 0);
       shift = - smin *scale;
   }

    cv::Mat out;


    if(mask.empty())
        img.convertTo(out, img.depth(), scale, shift );
    else
    {
        cv::Mat temp;
        img.convertTo(temp, img.depth(), scale, shift );
        temp.copyTo(out, mask);
    }

    CvMatMessage::Ptr img_out(new connection_types::CvMatMessage(enc::mono, stamp));
    img_out->value = out;
    msg::publish(out_image_, img_out);
}

void HistogramNormalization::process()
{
    CvMatMessage::ConstPtr msg = msg::getMessage<CvMatMessage>(in_mat_);

    if(!msg->hasChannels(1,CV_8U))  {
        throw std::runtime_error("image must be grayscale.");
    }

    CvMatMessage::ConstPtr mask_ptr;

    cv::Mat mask;

    if(msg::hasMessage(in_mask_)) {
        mask_ptr = msg::getMessage<CvMatMessage>(in_mask_);
        mask = mask_ptr->value;
    }

    normalize(msg->value, mask, msg->stamp_micro_seconds);
}




