// HEADER
#include "region_finder.h"

/// PROJECT
#include <csapex/msg/io.h>
#include <csapex/model/node_modifier.h>
#include <utils_param/parameter_factory.h>

/// SYSTEM
#include <csapex/utility/register_apex_plugin.h>
#include <boost/assign/std.hpp>


CSAPEX_REGISTER_CLASS(csapex::RegionFinder, csapex::Node)

using namespace csapex;
using namespace connection_types;


RegionFinder::RegionFinder()
{
    addParameter(param::ParameterFactory::declareRange("white tolerance", 0, 100, 60, 1));
    addParameter(param::ParameterFactory::declareRange("minimal width", 0, 400, 40, 10));
    addParameter(param::ParameterFactory::declareRange("minimal height", 0, 400, 30, 10));
    addParameter(param::ParameterFactory::declareRange("gap size", 0, 25, 15, 1));
    addParameter(param::ParameterFactory::declareBool("vertical gap", false));
}


void RegionFinder::setup(csapex::NodeModifier& node_modifier)
{
    in_mat_ = modifier_->addInput<CvMatMessage>("Original Image");

    out_ = modifier_->addOutput<VectorMessage, RoiMessage>("ROIs");
    out_debug_ = modifier_->addOutput<CvMatMessage>("debug");
}


void RegionFinder::findRegion(const cv::Mat &img,cv::Mat &debug, VectorMessage::Ptr out)
{
    int w_tol =readParameter<int>("white tolerance");
    int min_h = readParameter<int>("minimal height");
    int min_w = readParameter<int>("minimal width");
    int gap = readParameter<int>("gap size");
    bool v_gap = readParameter<bool>("vertical gap");

    int i,j;
    uchar* t;


    cv::Mat temp;
    if (img.channels() > 1)
    {
        cv::cvtColor(img, temp, CV_RGB2GRAY);
    } else {
        img.copyTo(temp);
    }

    std::vector<int> previous_row;
    previous_row.resize(temp.cols,0);
    std::vector<int> current_row;
    current_row.resize(temp.cols,0);
    std::vector<int> vertical_gap_counter;
            std::vector<int> vertical_gap_height;
    if (v_gap)
    {
        vertical_gap_counter.resize(temp.cols,0);
        vertical_gap_height.resize(temp.cols,0);
    }


    for( i = 0; i < temp.rows; ++i)
    {
        t = temp.ptr<uchar>(i);
        for (j = 0; j < temp.cols; ++j)
        {
            int upper = previous_row[j];
            int prev = 0;
            if (j!=0) {
                prev  = current_row[j-1];
            }

            if (t[j] < 255 - w_tol)
            { // the pixel doesn't fit in the range
                t[j] = 0;
                if (upper > 0) {       // there was a region active in the row above
                    if (v_gap)
                    {
                    vertical_gap_counter[j] = 1;
                    vertical_gap_height[j] = upper + 1;
                    }
                    if (prev > 0)
                    {
                        // the previous pixel contiuned the region already and the pixels above havn't been
                        // out of range for too long a time
                        current_row[j] = upper+1;
                        t[j] = 240;
                        continue;
                    } else {
                        current_row[j] = -1; // close-marker
                    }
                } else if (v_gap)
                {
                    if (vertical_gap_counter[j] > 0 && vertical_gap_counter[j] < gap)
                    {
                        vertical_gap_counter[j]++;
                        vertical_gap_height[j]++;
                    } else {
                        vertical_gap_counter[j] = 0;
                        vertical_gap_height[j] = 0;
                    }
                }
                if ((upper <= 0 || j == temp.cols - 1) && prev < 0)
                { // there wasn't a region active in the above row or this is the last column
                  // and the previous pixel is a close marker
                  // close region, get the width, reset current row to 0
                    int height = previous_row[j-1];
                    int width = -1;
                    int k = j-1;
                    if (j == temp.rows - 1)
                    {
                        height = std::max(previous_row[j-1], upper);
                        width++;
                    }
                    while (current_row[k] < 0 && k >= 0)
                    {
                        if (width < 0 && previous_row[k] < min_h/3)
                        {
                            // skip small "tails" at the end
                            k--;
                            continue;
                        }
                        else  if (previous_row[k] < min_h/3 )
                        {
                            // end before small tails at the beginning
                            break;
                        }
                        width++;
                        if (previous_row[k] > height)
                            height = previous_row[k];
                        current_row[k] = 0;
                        k--;
                    }

                    // if the region is big enough, add it to the output
                    if (height >= min_h && width >= min_w)
                    {
                        int start_x = k+1;
                        int start_y = i -height +1;
                        RoiMessage::Ptr roi(new RoiMessage);
                        cv::Scalar color(0,0,255);
                        roi->value = Roi(start_x, start_y, width, height, color);
                        out->value.push_back(roi);
                    }
                }
            } else { // the pixel does fit in the range 
                if (prev <= 0) {
                    // scan if there was a horizontal gap of allowed size
                    int k = 1;
                    bool g = false;
                    while (k < gap && k <= j)
                    {
                        if (current_row[j-k] > 0) { // a pixel within the colour range was found in the gap
                            g = true;
                            break;
                        }
                        k++;
                    }
                    if (g) { // if it was a gap, carry height-information over
                        int val = current_row[j-k];
                        k--;
                        while (k != -1){
                            current_row[j-k] = val;
                           // t[j-k] = 75;
                            k--;
                        }
                    } else { // no gap, this is a new region
                        if (upper <= 0) {
                            // check for vertical gaps
                            if (v_gap && vertical_gap_counter[j] > 0 && vertical_gap_counter[j] < gap)
                            {
                                current_row[j] = vertical_gap_height[j];
                            } else {
                                current_row[j] = 1;
                            }
                        } else {
                            current_row[j] = upper + 1;
                            // make regions rectangular for later roi-extraction
                            if (prev < 0) {
                                int k = j-1;
                                while (current_row[k] < 0 && k >= 0)
                                {
                                    current_row[k] = upper+1;
                                    k--;
                                    t[k] = 200;
                                }
                            }
                        }
                    }
                } else { // previous != 0
                    if (upper < 0)
                    {
                        current_row[j] = prev;
                    } else {
                        current_row[j] = upper + 1;
                    }
                }
                // colour the debug image
                t[j] = 255;
                // reset
                if (v_gap)
                {
                    vertical_gap_counter[j] = 0;
                    vertical_gap_height[j] = 0;
                }
            }
        } // inner loop end

        // close the active regions if the last row is reached
        /// TODO: Check minimum height before pixels are ignored
        if (i == temp.rows - 1)
        {
            for (int k = 0; k < temp.cols; k++)
            {
                if (current_row[k] != 0) {
                    int start_x = k;
                    int width = 0;
                    int height = 0;
                    if (current_row[k] < 0) {
                        height = previous_row[k];
                    }
                    /// TODO:
                    else if (current_row[k] > min_h/3){
                        height = current_row[k];
                     } else {
                        continue;
                     }
                    k++;
                    while ((current_row[k] < 0 || current_row[k] > min_h/3
                            ) && k < temp.cols)
                    {
                        width++;
                        if (current_row[k] < 0) {
                            if (previous_row[k] > height)
                                height = previous_row[k];
                        } else {
                            if (current_row[k] > height)
                                height = previous_row[k];
                        }
                        k++;
                    }
                    // if the region is big enough, add it to the output
                    if (height >= min_h && width >= min_w)
                    {
                        int start_y = i -height+1;
                        RoiMessage::Ptr roi(new RoiMessage);
                        cv::Scalar color(0,0,255);
                        roi->value = Roi(start_x, start_y, width, height, color);
                        out->value.push_back(roi);
                    }
                }
            }
        }

        // update current to previous
        previous_row = current_row;
        current_row.assign(temp.cols,0);
    } // outer loop end
     temp.copyTo(debug);
}




void RegionFinder::process()
{
    CvMatMessage::ConstPtr msg = msg::getMessage<CvMatMessage>(in_mat_);
    cv::Mat img = msg->value;

    CvMatMessage::Ptr debug(new CvMatMessage(enc::mono, msg->stamp_micro_seconds));
    cv::Mat de;

    VectorMessage::Ptr out(VectorMessage::make<RoiMessage>());

    findRegion(img,de, out);
    debug-> value = de;

    msg::publish(out_debug_, debug);
    msg::publish(out_, out);
}




