#include "target_finder.h"

/// PROJECT
#include <csapex/msg/io.h>
#include <csapex/model/node_modifier.h>
#include <csapex/utility/register_apex_plugin.h>
#include <sick14/target_message.h>
#include "../imagereprojector.h"

#include <utils_param/parameter_factory.h>
#include <utils_general/MathHelper.h>
#include <opencv2/opencv.hpp>

#include <algorithm>

CSAPEX_REGISTER_CLASS(csapex::TargetFinder, csapex::Node)

using namespace std;
using namespace csapex;
using namespace connection_types;

// ANSI color codes
const string RED = "\033[0;31m";
const string GREEN_BOLD = "\033[1;32m";
const string BLUE_BOLD = "\033[1;34m";
const string RST = "\033[0m";

TargetFinder::TargetFinder():
    has_camera_params_(false)
{
    addParameter(param::ParameterFactory::declareFileInputPath("transformation/config", ""),
                 boost::bind(&TargetFinder::updateConfig, this));
}


void csapex::TargetFinder::setup(csapex::NodeModifier& node_modifier)
{
    in_laser_      = modifier_->addInput<ScanMessage>("Front laser scan");
    in_target_pos_ = modifier_->addOptionalInput<CrossMessage>("Position of the target cross in the image.");
    input_transform_cam_base_link_ = modifier_->addInput<TransformMessage>("Transform (cam -> base_link)");
    input_transform_scan_cam_      = modifier_->addInput<TransformMessage>("Transform (scan -> cam)");


    out_target_pose_ = modifier_->addOutput<TargetMessage>("Target pose");
}

void csapex::TargetFinder::process()
{
    scan_ = msg::getMessage<ScanMessage>(in_laser_);

    if (msg::hasMessage(in_target_pos_)) {
        CrossMessage::ConstPtr cross = msg::getMessage<CrossMessage>(in_target_pos_);
        if (cross->type != CrossMessage::REGION_CENTER) {
            processTarget(cross);
        }
    }
}

void TargetFinder::processTarget(CrossMessage::ConstPtr target)
{
    try {
        TransformMessage::ConstPtr cam_to_base = msg::getMessage<TransformMessage>(input_transform_cam_base_link_);
        TransformMessage::ConstPtr scan_to_cam = msg::getMessage<TransformMessage>(input_transform_scan_cam_);

        tf::Vector3 pos_in_cam_frame;
        tf::Quaternion rot;

        // undistort
        std::vector<cv::Point2d> distorted;
        std::vector<cv::Point2d> undistorted;
        distorted.push_back(cv::Point2d(target->position.x, target->position.y));
        cv::undistortPoints(distorted, undistorted, camera_matrix_, distortion_coeffs_);
        double xu = undistorted[0].x;
        double yu = undistorted[0].y;

        // transform
        ImageReprojector::computePose(xu, yu, scan_->value, scan_to_cam->value, &pos_in_cam_frame, &rot);
        tf::Vector3 pos_in_child_frame = cam_to_base->value * pos_in_cam_frame;

        // fill and publish message
        TargetMessage::Ptr msg(new TargetMessage(cam_to_base->child_frame, target->stamp_micro_seconds));
        msg->transformation.value.setIdentity();
        msg->transformation.frame_id = cam_to_base->child_frame;
        msg->transformation.value.setOrigin(pos_in_child_frame);
        msg->transformation.value.setRotation(rot);

        switch (target->type) {
        case CrossMessage::INTERSECTION:
            msg->detector_type = TargetMessage::CROSS;
            //cout << GREEN_BOLD << "Cross" << RST << endl;
            break;
        case CrossMessage::LINE:
            msg->detector_type = TargetMessage::VERTICAL_LINE;
            //cout << BLUE_BOLD << "Line" << RST << endl;
            break;
        case CrossMessage::REGION_CENTER:
            msg->detector_type = TargetMessage::REGION_CENTER;
            //cout << RED << "Region" << RST << endl;
            break;
        }

        msg::publish(out_target_pose_, msg);
    }
    catch(...) {
        aerr << "Some error occured at target transformation." << std::endl;
    }
}

void TargetFinder::updateConfig()
{
    std::string path = readParameter<std::string>("transformation/config");
    ainfo << "read config " << path << std::endl;
    if(path.empty()) {
        return;
    }

    cv::FileStorage fs(path, cv::FileStorage::READ);
    if(!fs.isOpened()) {
        aerr << "cannot open config at " << path << std::endl;
        return;
    }
    fs["intrinsics"] >> camera_matrix_;
    fs["distortion"] >> distortion_coeffs_;
    fs.release();

    has_camera_params_ = true;
}
