#ifndef CROP_HPP
#define CROP_HPP
#include <opencv2/opencv.hpp>

namespace utils_cv {

inline void draw_rotated_crop(const cv::Mat &src, cv::Mat &dst, const cv::Point2f &center,
                              const cv::Size2f &size, const double angle, const double scale)
{
    cv::RotatedRect rotation_roi(center, size, angle);
    cv::Rect        rotation_roi_bounding(rotation_roi.boundingRect());


    /// check if rectangle is in image
    if(rotation_roi_bounding.x < 0 || rotation_roi_bounding.x + rotation_roi_bounding.width >= src.cols)
        return;
    if(rotation_roi_bounding.y < 0 || rotation_roi_bounding.y + rotation_roi_bounding.height >= src.rows)
        return;

    cv::Mat rotation_roi_img(src, rotation_roi_bounding);

    std::vector<cv::Point2f> points(4);
    cvBoxPoints(rotation_roi, (CvPoint2D32f*) points.data());
    cv::Point2f corner(rotation_roi_bounding.x, rotation_roi_bounding.y);
    cv::line(rotation_roi_img, points[0] - corner,  points[1] - corner, cv::Scalar(255,0,0));
    cv::line(rotation_roi_img, points[1] - corner,  points[2] - corner, cv::Scalar(255,0,0));
    cv::line(rotation_roi_img, points[2] - corner,  points[3] - corner, cv::Scalar(255,0,0));
    cv::line(rotation_roi_img, points[3] - corner,  points[0] - corner, cv::Scalar(255,0,0));
    dst = rotation_roi_img.clone();
}

inline cv::Mat getRotatedCrop(const cv::Mat& src, cv::Point2f pos, cv::Size size, float angle, float scale) {
    pos.x = (int)pos.x;
    pos.y = (int)pos.y;
    const cv::RotatedRect rect(pos, size, angle);

    // create bounding box of rotated and unrotated rectangle
    cv::Size2f boundingBoxSize(std::max((int)(rect.boundingRect().width/scale+0.5f),  size.width),
                           std::max((int)(rect.boundingRect().height/scale+0.5f), size.height));
    cv::Rect boundingBox = cv::RotatedRect(pos, boundingBoxSize, 0).boundingRect();

    //Point a cv::Mat header at it (no allocation is done)
    cv::Mat cropped(src, boundingBox);

    // center relative to cropped region
    cv::Point2f boundingBoxCenter(boundingBox.width/2.0f, boundingBox.height/2.0f);

    // get the rotation matrix
    cv::Mat M = getRotationMatrix2D(boundingBoxCenter, rect.angle, scale);

    // perform the affine transformation
    cv::Mat rotated;
    cv::warpAffine(cropped, rotated, M, cv::Size(cropped.cols, cropped.rows), cv::INTER_NEAREST);

    // crop the resulting image
    cv::Rect outRect = cv::RotatedRect(boundingBoxCenter, cv::Size2f(size.width-1, size.height-1), 0).boundingRect();
    cv::Mat cropped2(rotated, outRect);

    return cropped2;
}

inline double compare(const cv::Mat &mat1, const cv::Mat &mat2, const double norm = 0)
{
    assert(mat1.rows == mat2.rows);
    assert(mat1.cols == mat2.cols);
    cv::Mat zeros;
    //    cv::subtract(mat1, mat2, zeros);
    cv::bitwise_xor(mat1, mat2, zeros);
    int non_zero = cv::countNonZero(zeros);
    int all = mat1.rows * mat1.cols;
    if(norm != 0)
        return (all - non_zero) / norm;
    return (all - non_zero) / all;
}

inline void crop(const cv::Mat &src, cv::Mat &dst, const cv::Point2f &center,
                 const cv::Size2f &size, const double angle, const double scale)
{
//    cv::Mat mutable_src = src;
    cv::RotatedRect roi_rect(center, size, angle);
    cv::Rect crop_rect(roi_rect.boundingRect());
    cv::Mat roi(src, crop_rect);
//    cv::rectangle(mutable_src, crop_rect, cv::Scalar::all(0));
    cv::Point2f roi_center(roi.cols / 2.0, roi.rows / 2.0);
    cv::Mat rotation = cv::getRotationMatrix2D(roi_center, angle, scale);
    cv::warpAffine(roi, dst, rotation, crop_rect.size(), cv::INTER_NEAREST);

    std::vector<cv::Mat> channels;
    cv::split(dst, channels);

    for(int i = 0; i < channels.size(); ++i) {
        cv::getRectSubPix(channels[i], size, roi_center, channels[i]);
    }

    cv::merge(channels, dst);
}

/**
 * ! Supporting crops till 4 channels.
 */
inline void crop2(const cv::Mat &src, cv::Mat &dst, const cv::Point2f &center,
                  const cv::Size2f &size, const double angle, const double scale)
{
    /// Adjust the scale
    cv::Size2f scaled_size = size;
    scaled_size.height *= scale;
    scaled_size.width  *= scale;

    /// Calculate the working boundaries
    cv::RotatedRect rotation_roi(center, scaled_size, angle);
    cv::Rect        rotation_roi_bounding(rotation_roi.boundingRect());

    /// check if rectangle is in image
    if(rotation_roi_bounding.x < 0 || rotation_roi_bounding.x + rotation_roi_bounding.width >= src.cols)
        return;
    if(rotation_roi_bounding.y < 0 || rotation_roi_bounding.y + rotation_roi_bounding.height >= src.rows)
        return;

    cv::Mat rotation_roi_img(src, rotation_roi_bounding);
//    /// DEBUG
//    draw_rotated_crop(src, rotation_roi_img, center, scaled_size, angle, scale);
//    cv::imshow("rotation roi", rotation_roi_img);
//    cv::waitKey(0);
//    /// GUBED

    cv::Point2f rotation_center(rotation_roi_img.rows / 2.0, rotation_roi_img.cols / 2.0);
    cv::Mat     rotation = cv::getRotationMatrix2D(rotation_center, angle, 1.0);
    cv::warpAffine(rotation_roi_img, rotation_roi_img, rotation, cv::Size(rotation_roi_img.cols, rotation_roi_img.rows),
                   cv::INTER_NEAREST, cv::BORDER_CONSTANT, cv::Scalar::all(255));
    cv::getRectSubPix(rotation_roi_img, scaled_size, rotation_center, dst);
    cv::resize(dst, dst, size, 0, 0, cv::INTER_NEAREST);
}

inline void rotate(const cv::Mat &src, cv::Mat& dst, const double angle, const cv::Scalar &def = cv::Scalar::all(255))
{
    int size = std::sqrt(src.cols * src.cols + src.rows * src.rows);
    cv::Mat canvas(size, size, src.type(), def);

    cv::Mat canvas_roi(canvas, cv::Rect((size - src.cols) / 2.0, (size - src.rows) / 2.0, src.cols, src.rows));
    src.copyTo(canvas_roi);

    cv::Point2f     center(canvas.cols * 0.5, canvas.rows * 0.5);
    cv::Mat         rot = cv::getRotationMatrix2D(center, angle, 1.0);
    cv::warpAffine(canvas, canvas, rot, cv::Size(canvas.cols, canvas.rows), cv::INTER_NEAREST, cv::BORDER_CONSTANT, def);

    cv::RotatedRect rotRec(center, cv::Size2f(src.cols, src.rows), angle);

    cv::Rect toCopy = rotRec.boundingRect();
    toCopy.width = std::min(canvas.cols, toCopy.width - 1);
    toCopy.height = std::min(canvas.rows, toCopy.height  - 1);

    dst = canvas(toCopy);
    return;

}
}

#endif // CROP_HPP

